#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Command line interface set up."""

__metaclass__ = type
__all__ = [
    'get_parser',
    'main',
    'prepare_parser',
    'run',
    ]

import sys

from lpsetup import (
    __doc__ as description,
    argparser,
    exceptions,
    )
from lpsetup.subcommands import (
    init_target,
    initlxc,
    initrepo,
    install_lxc,
    update,
    version,
    )
from lpsetup.utils import confirm


# The global help epilog, built collecting subcommands' epilogs,
# reflects the order of the following sequence.
SUBCOMMANDS = (
    ('init-lxc', initlxc.SubCommand()),
    ('init-target', init_target.SubCommand()),
    ('init-repo', initrepo.SubCommand()),
    ('update', update.SubCommand()),
    ('install-lxc', install_lxc.SubCommand()),
    ('version', version.SubCommand()),
    )


def run(parser, subcommand, namespace):
    """Run the *subcommand* using the given *namespace*.

    This function takes care of::

        - namespace initialization and validation
        - interactive arguments handling
        - executing a dry run, if requested
        - restarting the same subcommand as root, if required
        - actual subcommand execution
    """
    # Initialize the namespace.
    argparser.init_namespace(namespace)

    # Prepare the namespace. Each subcommand can define a *prepare_namespace*
    # method that can raise a ValidationError if the namespace is not valid.
    try:
        subcommand.prepare_namespace(namespace)
    except exceptions.ValidationError as err:
        parser.error(err)
    except AttributeError:
        pass

    # Handle user interaction and dry run.
    if namespace.dry_run or namespace.interactive:
        # Display the subcommand description.
        command_description = subcommand.get_description(namespace)
        if command_description:
            print 'This command will perform the following actions:\n'
            print command_description + '\n'
        # Quit without errors if this is a dry run.
        if namespace.dry_run:
            return
        # If this is not a dry run, then it is an interactive one.
        # Prompt the user for confirmation to proceed and quit if
        # requested (in this case with exit code 1).
        try:
            proceed = confirm()
        except KeyboardInterrupt:
            print '\nQuitting.'
            return 1
        if not proceed:
            return 1

    # Restart as root if needed.
    if not namespace.run_as_root and getattr(
        subcommand, 'root_required', False):
        return argparser.restart_as_root(parser, namespace)

    # Execute the command.
    try:
        return subcommand.run(namespace)
    except exceptions.ExecutionError as err:
        return err


def prepare_parser(parser, subcommand):
    """Add common and subcommand specific arguments to the given *parser*.

    Also add to the parser defaults a *main* function that, taking a parsed
    namespace, executes the current subcommand with the current parser.
    """
    def main(namespace):
        return run(parser, subcommand, namespace)

    # Store a reference to the current subcommand entry point.
    parser.set_defaults(main=main)
    # Add user interaction related arguments.
    argparser.add_common_arguments(
        parser,
        has_dry_run=getattr(subcommand, 'has_dry_run', False),
        has_interactive_run=getattr(subcommand, 'has_interactive_run', False),
        )
    # Add subcommand specific arguments, if requested.
    try:
        subcommand.add_arguments(parser)
    except AttributeError:
        pass


def get_parser(subcommands):
    """Return the argument parser with all subcommands registered.

    Also add the help subcommand and the global epilog, which is generated
    by collecting the epilogs from all of the subcommands.
    """
    parser = argparser.ArgumentParser(description=description)
    subparsers = parser.add_subparsers(
        title='positional arguments', metavar='subcommand',
        help='One of the following subcommands. Each subcommand accepts '
             '-h or --help to describe it.')
    epilogs = ['%(prog)s provides several subcommands that can be used to '
               'create and update Launchpad development and testing '
               'environments. Examples of how to use the available '
               'subcommands follow.']
    for name, subcommand in subcommands:
        # Prepare the subcommand help and epilog.
        help = subcommand.help
        epilog = getattr(subcommand, 'epilog', None)
        if epilog is not None:
            # Retrieve the program name including base name and subcommand.
            prog = '{0} {1}'.format(parser.prog, name)
            epilogs.extend([
                '=== {0} ==='.format(prog),
                epilog % {'prog': prog}
                ])
        # Register the subcommand.
        subparser = subparsers.add_parser(
            name, description=help, help=help, epilog=epilog)
        # Prepare the subcommand arguments.
        prepare_parser(subparser, subcommand)

    # Add the global epilog.
    parser.epilog = '\n\n'.join(epilogs)

    # Add the *help* subcommand.
    def main(namespace):
        command = namespace.command
        help = parser.prefix_chars + 'h'
        args = [help] if command is None else [command, help]
        parser.parse_args(args)

    choices = dict(subcommands).keys()
    help = 'More help on a command.'
    subparser = subparsers.add_parser('help', description=help, help=help)
    subparser.add_argument('command', nargs='?', choices=choices)
    subparser.set_defaults(main=main)

    return parser


def main(args=None, parser=None):
    # Retrieve the command line arguments.
    if args is None:
        args = sys.argv[1:]
    # Get the parser.
    if parser is None:
        parser = get_parser(SUBCOMMANDS)
    # Parse the command line arguments.
    try:
        namespace = parser.parse_args(args)
        # Run the subcommand. Each subparser adds to namespace a main
        # function that basically execute run() passing the corresponding
        # subcommand, sub parser and the namespace itself.
        return namespace.main(namespace)
    except SystemExit as err:
        return err.code
