#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""init-target subcommand: prepare a target machine to run Launchpad."""

__metaclass__ = type
__all__ = [
    'initialize',
    'initialize_base',
    'initialize_lxc',
    'setup_apt',
    'setup_home',
    'SubCommand',
    ]

from email.Utils import formataddr
import os
import pwd
import shutil
import subprocess

from shelltoolbox import (
    apt_get_install,
    file_append,
    generate_ssh_keys,
    mkdirs,
    run,
    su,
    user_exists,
    )

from lpsetup import (
    argparser,
    get_version,
)
from lpsetup.handlers import (
    handle_lpuser_as_username,
    handle_ssh_keys,
    handle_user,
    handle_userdata,
    )
from lpsetup.settings import (
    APT_REPOSITORIES,
    BASE_PACKAGES,
    HOSTS_CONTENT,
    HOSTS_FILE,
    LP_APACHE_MODULES,
    LP_APACHE_ROOTS,
    LP_PACKAGES,
    SSH_KEY_NAME,
    )
from lpsetup.utils import (
    call,
    render_to_file,
    running_in_container,
    wrap_contents,
    )

NEWLINE = '\n'


def get_version_path():
    return '/var/lib/lpsetup/' + get_version()


def get_backup_path():
    return os.path.join(get_version_path(), 'originals')


def make_version_dir(mkdirs=mkdirs):
    mkdirs(get_version_path())
    mkdirs(get_backup_path())


def make_backup(source, get_backup_path=get_backup_path, mkdirs=mkdirs):
    """Make a backup of a file in the current version's backup location."""
    # We don't use os.path.join here because both of the input paths are
    # absolute (i.e., they start with a slash) so we instead construct the
    # path ourselves in order to "re-root" the otherwise absolute source path
    # in the backup directory.
    destination = os.path.normpath(get_backup_path() + os.path.sep + source)
    destination_path = os.path.dirname(destination)
    mkdirs(destination_path)
    shutil.copyfile(source, destination)


def write_file_contents(filename, contents, mode, header=None):
    """Write `contents` to filename and chmod it to 0644."""
    assert mode in ('w', 'a'), "Mode should be 'w' or 'a' only."
    if mode == 'w' and os.path.exists(filename):
        with open(filename, 'r') as existing_file:
            file_contents = existing_file.read()
        if file_contents.strip() != contents.strip():
            raise Exception(
                "File %s already exists with different contents." % filename)
        else:
            return
    with open(filename, mode) as f:
        if header is not None:
            f.write(header + '\n')
        f.write(contents)
    os.chmod(filename, 0644)


def setup_ssh(ssh_dir, valid_ssh_keys, ssh_key_path):
    """Set up the user's `.ssh` directory.

    Also ensure that SSH is configured correctly for the user.
    """
    # Set up the user's ssh directory.  The ssh key must be associated
    # with the lpuser's Launchpad account.
    mkdirs(ssh_dir)
    # Generate user ssh keys if they don't exist.
    pub_key_path = ssh_key_path + '.pub'
    if not valid_ssh_keys:
        generate_ssh_keys(ssh_key_path)
    # Write the user's public key to `authorized_keys` so that the user can
    # non-interactively ssh to the container.
    authorized_keys = os.path.join(ssh_dir, 'authorized_keys')
    authorized_keys_contents = open(pub_key_path).read()
    # Add bazaar.launchpad.net to `known_hosts`.
    known_hosts = os.path.join(ssh_dir, 'known_hosts')
    known_host_contents = run(
        'ssh-keyscan', '-t', 'rsa', 'bazaar.launchpad.net')
    files_to_write = [
        (authorized_keys, authorized_keys_contents, 'a'),
        (known_hosts, known_host_contents, 'a'),
        ]
    for filename, contents, mode in files_to_write:
        wrapped = wrap_contents(contents)
        write_file_contents(filename, wrapped, mode)


def initialize_base(user):
    """Perform base initialization shared by init-target and init-lxc."""
    # Install necessary deb packages.  This requires Oneiric or later.
    call('apt-get', 'update')
    apt_get_install(*BASE_PACKAGES, caller=call)
    # Create the user (if he does not exist).
    if not user_exists(user):
        call('useradd', '-m', '-s', '/bin/bash', '-U', user)

initialize_base.description = """Update your system and install necessary \
    deb packages ({packages}).
    Create the user $user if it does not exist.
""".format(packages=' '.join(BASE_PACKAGES))


def initialize(user):
    """Initialize target machine."""
    make_version_dir()
    initialize_base(user)
    with su(user):
        # Create Apache document roots, to avoid warnings.
        mkdirs(*LP_APACHE_ROOTS)

    # Set up Apache modules.
    for module in LP_APACHE_MODULES.split():
        call('a2enmod', module)

    # The user must be added to sudoers if init-target is run in a container
    # and we are running the developer story.
    # XXX 2012-07-18 frankban: add the developer/testing check.
    if running_in_container():
        subprocess.call(['adduser', user, 'sudo'])
        # Add the user to his own group.
        pwd_database = pwd.getpwnam(user)
        subprocess.call(['addgroup', '--gid', str(pwd_database.pw_gid), user])

    # Set up container hosts file.
    make_backup(HOSTS_FILE)
    lines = ['{0}\t{1}\n'.format(ip, names)
             for ip, names in HOSTS_CONTENT]
    contents = wrap_contents(NEWLINE.join(lines))
    for line in contents.split(NEWLINE):
        file_append(HOSTS_FILE, line)

initialize.description = initialize_base.description + """ Create Apache \
    document roots for launchpad and enable required Apache \
    modules ({modules}).
    Set up hosts file for Launchpad ({hosts}).
""".format(modules=LP_APACHE_MODULES, hosts=HOSTS_FILE)


def setup_home(user, full_name, email, lpuser, valid_ssh_keys, ssh_key_path):
    """Initialize the user home directory.

    This is a separate step for several reasons::

        - It is directly called by *init-lxc* as one of its first steps.
          Note that *init-lxc* needs the user's `.ssh` dir to be set up
          to be able to connect to the container.
        - *init-lxc* should be able to skip this step when re-executing
          *init-target* from inside the container, to avoid setting up the
          user home two times. Note: the user home is bind mounted by lxc.
    """
    with su(user) as env:
        # Set up the `.ssh` directory.
        setup_ssh(os.path.join(env.home, '.ssh'), valid_ssh_keys, ssh_key_path)

        # Set up bzr and Launchpad authentication.
        call('bzr', 'whoami', formataddr([full_name, email]))
        if valid_ssh_keys:
            subprocess.call(['bzr', 'launchpad-login', lpuser])

setup_home.description = """Set up the user's ssh directory ($home_dir/.ssh).
    Create, if it does not exist, the ssh key $ssh_key_path and authorize \
    this key for the user $user.
    Add bazaar.launchpad.net to known hosts.
    Set up bazaar authentication: $full_name <$email>.
    Set up Launchpad user id: $lpuser.
"""


def initialize_lxc():
    """Initialize LXC container.

    Note that this step is guarded by a filter callable method so that
    it is only called when lpsetup is running in a container.
    """
    lxc_os = get_distro()
    # XXX benji 2012-03-19 bug=959352: this is so graphviz will work in an
    # ephemeral container
    mkdirs('/rootfs/usr/lib')
    dst = '/rootfs/usr/lib/graphviz'
    if not os.path.exists(dst):
        os.symlink('/usr/lib/graphviz', dst)
    # XXX gary 2012-06-26 bug=1014916: this fixes a bug in Lucid LXC
    # containers.
    if lxc_os == 'lucid':
        # Change /var/lib/lxc/lptests/rootfs/etc/init/networking.conf to not
        # wait for udev (that is, use start on (local-filesystems)).
        udevtrigger = '/etc/init/udevtrigger.conf'
        if os.path.exists(udevtrigger):
            make_backup(udevtrigger)
            os.remove(udevtrigger)
        networking = '/etc/init/networking.conf'
        make_backup(networking)
        render_to_file('networking.conf', {}, networking)

initialize_lxc.description = None


def get_distro():
    return run('lsb_release', '-cs').strip()


def setup_apt():
    """Setup, update and upgrade deb packages."""
    distro = get_distro()
    # APT repository update.
    for repository in APT_REPOSITORIES:
        assume_yes = None if distro == 'lucid' else '-y'
        call('add-apt-repository', assume_yes,
             repository.format(distro=distro))
    call('apt-get', 'update')
    # Install base and Launchpad deb packages.
    apt_get_install(*LP_PACKAGES, LC_ALL='C', caller=call)

setup_apt.description = """Add required APT repositories and install \
    Launchpad dependencies: {dependencies}.
""".format(dependencies=' '.join(LP_PACKAGES))


def setup_database(user):
    """Setup Launchpad database."""
    with su(user):
        script_name = 'launchpad-database-setup.sh'
        here = os.path.dirname(__file__)
        script_path = os.path.join(here, '..', 'bin', script_name)
        with open(script_path, 'r') as f:
            line = f.readline()
            interpreter = '/bin/sh'
            assert line.startswith('#!' + interpreter), (
                'Interpreter for {0} is not {1}.'.format(
                    script_path, interpreter))
        call(interpreter, script_path, user)

setup_database.description = """Setup database for Launchpad use."""


class SubCommand(argparser.StepsBasedSubCommand):
    """Prepare a machine to run Launchpad.  May be an LXC container or not."""
    initialize_step = (initialize, ['user'])

    setup_home_step = (
        setup_home, ['user', 'full_name', 'email', 'lpuser',
            'valid_ssh_keys', 'ssh_key_path'])

    initialize_lxc_step = (
        initialize_lxc, [], lambda namespace: running_in_container())

    setup_apt_step = (setup_apt, [])
    setup_database_step = (setup_database, ['user'])

    steps = (
        initialize_step,
        setup_home_step,
        initialize_lxc_step,
        setup_apt_step,
        setup_database_step,
        )

    help = __doc__
    root_required = True
    handlers = (
        handle_user,
        handle_lpuser_as_username,
        handle_userdata,
        handle_ssh_keys,
        )

    epilog = """`%(prog)s` prepares a Launchpad environment in the current \
        machine, setting up ssh keys if needed and initializing Launchpad \
        dependencies. The machine can either be your local system, an LXC \
        container or a cloud instance. However, please refer to \
        `lp-setup init-lxc` for an more convenient way to set up a Launchpad \
        environment inside an LXC.

        Prepare a machine to run Launchpad:

        \t$ %(prog)s

        By default the home of the current user is configured, but if this \
        subcommand is run as root, you need to explicitly specify the user:

        \t# %(prog)s -u myuser

        In this case, if `myuser` does not exist, it will be created.

        The current user (or the provided one) should have a correctly \
        configured bazaar user id (i.e `bzr whoami`). However, sometimes this \
        is not possible (e.g. when the provided user does not exist). In this \
        case, you can explicitly pass the user's full name and email and \
        the bazaar user id will be set up using provided arguments:

        \t# %(prog)s -u myuser -f 'My User' -e myuser@example.com
    """

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        parser.add_argument(
            '-u', '--user',
            help='The name of the system user to be created or updated. '
                 'The current user is used if this script is not run as '
                 'root and this argument is omitted.')
        parser.add_argument(
            '-E', '--email',
            help='The email of the user, used for bzr whoami. This argument '
                 'can be omitted if a bzr id exists for current user.')
        parser.add_argument(
            '-f', '--full-name',
            help='The full name of the user, used for bzr whoami. '
                 'This argument can be omitted if a bzr id exists for '
                 'current user.')
        parser.add_argument(
            '-l', '--lpuser',
            help='The name of the Launchpad user that will be used to '
                 'check out dependencies. If not provided, the system '
                 'user name is used.')
        parser.add_argument(
            '-S', '--ssh-key-name', default=SSH_KEY_NAME,
            help='The ssh key name used to connect to Launchpad. '
                 '[DEFAULT=%(default)s]')
