#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""init-repo subcommand: prepare source code destinations and download them.

Prepares the user's Launchpad directory structure and downloads the
latest Launchpad source code.  Sets up ~/.bazaar/locations.conf
appropriately.  The repository uses the sandbox development model as
specified in `lxc-install`.

Run as the user in the Launchpad target, either a host machine or LXC
container.
"""

__metaclass__ = type
__all__ = [
    'fetch',
    'setup_bzr_locations',
    'SubCommand',
    ]

import os
from StringIO import StringIO
import subprocess

from shelltoolbox import mkdirs

from lpsetup import (
    argparser,
    exceptions,
    handlers,
    )
from lpsetup.settings import (
    LP_BRANCH_NAME,
    LP_BZR_LOCATIONS,
    LP_CHECKOUT_NAME,
    LP_HTTP_REPO,
    LP_REPOSITORY_DIR,
    LP_SSH_REPO,
    )
from lpsetup.utils import (
    call,
    ConfigParser,
    wrap_contents,
    )


def fetch(source, repository, branch_name, checkout_name, no_checkout):
    """Create a repo for the Launchpad code and retrieve it."""
    # XXX should we warn if root?
    # Set up the repository.
    mkdirs(repository)
    is_lightweight = not no_checkout
    no_trees = '--no-trees' if is_lightweight else None
    # Initialize the repository.

    # Skip creating the repository if it already exists.  There is the
    # risk that it is not made with the same `no_trees` option as
    # specified here.
    if not os.path.exists(os.path.join(repository, '.bzr')):
        try:
            call('bzr', 'init-repo', '--quiet', repository, no_trees)
        except subprocess.CalledProcessError as err:
            msg = 'Error: unable to initialize the repository: '
            raise exceptions.ExecutionError(msg + err.output)
    # Set up the codebase.
    branch_dir = os.path.join(repository, branch_name)
    # Retrieve the branch.
    if os.path.exists(branch_dir):
        # The branch already exists, so we pull to refresh it rather than
        # creating it.
        cmd = ['bzr', 'pull', '--overwrite', '-d', branch_dir, source]
    else:
        cmd = ['bzr', 'branch', source, branch_dir]
    try:
        call(*cmd)
    except subprocess.CalledProcessError as err:
        msg = 'Error: unable to branch source: '
        raise exceptions.ExecutionError(msg + err.output)
    if is_lightweight:
        checkout_dir = os.path.join(repository, checkout_name)
        # Create a lightweight checkout if it doesn't exist.
        if not os.path.exists(checkout_dir):
            try:
                call('bzr', 'co', '--lightweight', branch_dir, checkout_dir)
            except subprocess.CalledProcessError as err:
                msg = 'Error: unable to create the lightweight checkout: '
                raise exceptions.ExecutionError(msg + err.output)

fetch.description = """Set up a Launchpad repository in $repository, \
    retrieving the source code from $source.
"""


def setup_bzr_locations(
    lpuser, repository, branch_name, template=LP_BZR_LOCATIONS):
    """Set up bazaar locations.

    Note that this step is guarded by a filter callable so that it is only
    called when the user Launchpad login can be retrieved.
    """
    context = {
        'branch_dir': os.path.join(repository, branch_name),
        'repository': repository,
        'lpuser': lpuser,
        }

    bazaar_dir = os.path.expanduser('~/.bazaar')
    mkdirs(bazaar_dir)
    path = os.path.join(bazaar_dir, 'locations.conf')
    parser = ConfigParser()
    parser.read(path)
    for section_template, options in template.items():
        section = section_template.format(**context)
        if not parser.has_section(section):
            parser.add_section(section)
        for option, value in options.items():
            parser.set(section, option, value.format(**context))
        with open(path, 'w') as f:
            contents = StringIO()
            parser.write(contents)
            f.write(wrap_contents(contents.getvalue()))
            contents.close()

setup_bzr_locations.description = """Update bazaar locations \
    ($home_dir/.bazaar/locations.conf) to include repository \
    $repository and branch $branch_name.
"""


class SubCommand(argparser.StepsBasedSubCommand):
    """Get the Launchpad code and dependent source code."""

    steps = (
        (fetch,
         ['source', 'repository', 'branch_name', 'checkout_name',
            'no_checkout']),
        (setup_bzr_locations, ['lpuser', 'repository', 'branch_name'],
         lambda namespace: namespace.lpuser is not None),
        )

    help = __doc__
    handlers = (
        handlers.handle_user,
        handlers.handle_lpuser_from_lplogin,
        handlers.handle_directories,
        handlers.handle_branch_and_checkout,
        handlers.handle_source,
        )
    root_required = False

    epilog = """Once a Launchpad environment is correctly initialized \
        (i.e. using `init-lxc` or `init-target`) `%(prog)s` can be used \
        to retrieve the Launchpad code and its dependencies.

        By default, the code is retrieved from `{ssh_repo}`, a shared \
        repository without trees is created in `{repository}`, containing a \
        lightweight checkout named `{checkout}`:

        \t$ %(prog)s

        However, it is possible to change these parameters, e.g. to create a \
        repository in `~/mylaunchpad` containing a lightweight checkout \
        `mycheckout` and retrieving code from `lp:~myuser/launchpad/mybranch`:

        \t$ %(prog)s --source lp:~myuser/launchpad/mybranch -r ~/mylaunchpad \
        --checkout-name mycheckout

        As seen above, the default behavior is to get the code using bzr+ssh: \
        we assume you set up a valid Launchpad user id. This way you can \
        contribute to Launchpad, commit code, fix bugs, etc. Anyway, \
        sometimes an http checkout can be more convenient (e.g. when creating \
        testing environments). To retrieve Launchpad from `{http_repo}`:

        \t$ %(prog)s --use-http

        Lightweight checkouts are considered an effective way of working with \
        bazaar repositories. On the other hand, developers may prefer other \
        approaches (normal branches with trees, co-located branches). It is \
        possible to initialize a Launchpad repository with trees as follow:

        \t$ %(prog)s --no-checkout --branch-name trunk

        If `--branch-name` is not provided `{branch}` is used by default.
    """.format(
        branch=LP_BRANCH_NAME, checkout=LP_CHECKOUT_NAME,
        http_repo=LP_HTTP_REPO, repository=LP_REPOSITORY_DIR,
        ssh_repo=LP_SSH_REPO)

    @staticmethod
    def add_common_arguments(parser):
        parser.add_argument(
            '--source', default=None,
            help='Location from which to retrieve Launchpad source.  Default '
                'is lp:launchpad (if --use-http is false), otherwise '
                'http://bazaar.launchpad.net/~launchpad-pqm/launchpad/devel.')
        parser.add_argument(
            '--use-http', action='store_true', dest='use_http',
            default=False, help='Use http in place of bzr+ssh, ignored if '
            '--source is provided.  Defaults to off.')
        parser.add_argument(
            '--branch-name', default=LP_BRANCH_NAME,
            help='Name of the directory in which to place the branch. '
                 '[DEFAULT=%(default)s]')
        parser.add_argument(
            '--checkout-name', default=LP_CHECKOUT_NAME,
            help='Create a checkout with the given name. Ignored if '
                 '--no-checkout is specified. [DEFAULT=%(default)s]')
        parser.add_argument(
            '-r', '--repository', default=LP_REPOSITORY_DIR,
            help='The directory of the Launchpad repository to be created. '
                 'The directory must reside under the home directory of the '
                 'given user (see -u argument). [DEFAULT=%(default)s]')
        parser.add_argument(
            '--no-checkout', action='store_true', dest='no_checkout',
            default=False,
            help='Initialize a bzr repository with trees and '
                 'do not create a lightweight checkout.')

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        self.add_common_arguments(parser)
