#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Update subcommand: updates a Launchpad development environment."""

__metaclass__ = type
__all__ = [
    'initialize_directories',
    'SubCommand',
    'update_dependencies',
    'update_tree',
    ]

import os.path

from shelltoolbox import (
    cd,
    get_su_command,
    mkdirs,
    )

from lpsetup import argparser
from lpsetup import handlers
from lpsetup.settings import LP_SOURCE_DEPS
from lpsetup.utils import call


def initialize_directories(target_dir, external_path):
    """Initialize the eggs, yui, and sourcecode directories.

    Create them if necessary.
    """
    for dir_ in ['eggs', 'yui', 'sourcecode']:
        mkdirs(os.path.join(target_dir, external_path, dir_))

initialize_directories.description = """Initialize the eggs, yui, and \
    sourcecode directories in $target_dir.
"""


def update_dependencies(target_dir, external_path, use_http, lp_source_deps):
    """Update the external dependencies."""
    use_http_param = '--use-http' if use_http else None
    cmd = os.path.join(target_dir, 'utilities', 'update-sourcecode')
    abs_external_path = os.path.abspath(
        os.path.join(target_dir, external_path))
    source_path = os.path.join(abs_external_path, 'sourcecode')
    call(cmd, use_http_param, source_path)

    # Update the download cache.
    download_cache = os.path.join(target_dir, 'download-cache')
    if os.path.exists(download_cache):
        call('bzr', 'up', download_cache)
    else:
        call('bzr', 'co', '-v', '--lightweight',
             lp_source_deps, download_cache)

    # Link to the external sourcecode.
    if abs_external_path != target_dir:
        cmd = os.path.join(
            target_dir, 'utilities', 'link-external-sourcecode')
        call(cmd,
            '--target', target_dir,
            '--parent', external_path)

update_dependencies.description = """Update Launchpad external \
    dependencies in $target_dir.
"""


def update_tree(target_dir):
    """Update the tree at target_dir with the latest LP code."""
    with cd(target_dir):
        call('bzr', 'pull')

update_tree.description = """Update the tree at $target_dir with the \
    latest LP code.
"""


def make_schema(user, target_dir):
    """Set up the Launchpad environment."""

    # Make and install launchpad.
    with cd(target_dir):
        # Using real su because mailman make script uses uid.
        call(*get_su_command(user, ['make', 'schema']))

make_schema.description = """Run 'make schema' within the Launchpad tree."""


class SubCommand(argparser.StepsBasedSubCommand):
    """Updates an existing Launchpad development environment.

    Gets new versions of Launchpad source and external sources.
    """

    has_interactive_run = False
    steps = (
        (initialize_directories, ['target_dir', 'external_path']),
        (update_dependencies,
           ['target_dir', 'external_path', 'use_http', 'lp_source_deps']),
        (update_tree, ['target_dir']),
        (make_schema, ['user', 'target_dir']),
        )
    help = __doc__
    handlers = (
        # Normalize paths and default to cwd if none exists.
        handlers.handle_user,
        handlers.handle_target_dir,
        )

    epilog = """`%(prog)s` updates an existing Launchpad branch, contextually \
        retrieving the current external sources and dependencies.

        To update a branch in the current working directory:

        \tcd ~/launchpad/lp-branches/mybranch
        \t%(prog)s

        To update a branch specifying its path:

        \t%(prog)s ~/launchpad/lp-branches/mybranch

        By default this subcommand updates the download cache checking out \
        {deps}. To change the source repository for the cache, use \
        `--lp-source-deps`:

        \t%(prog)s --lp-source-deps my-source-deps
    """.format(deps=LP_SOURCE_DEPS)

    @staticmethod
    def add_common_arguments(parser):
        parser.add_argument(
            '-e', '--external-path', default='.',
            help='Path to directory that contains sourcecode '
                 'and download-cache directories, relative to the '
                 'working-dir. ')
        parser.add_argument(
            '--lp-source-deps', default=LP_SOURCE_DEPS,
            help='URL of LP source deps branch. [DEFAULT=%(default)s]')

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        parser.add_argument(
            'target_dir', nargs='?', default=os.getcwd(),
            help='Path to branch to update. [DEFAULT=current directory]')
        self.add_common_arguments(parser)
        parser.add_argument(
            '--use-http', default=False, action='store_true',
            help='Force bzr to use http to get the sourcecode '
                 'branches rather than using bzr+ssh.')
