#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Common resources for integration tests."""

__metaclass__ = type
__all__ = [
    'IntegrationTestBase',
    ]

import argparse
from datetime import datetime
from email.Utils import formataddr
import os
import sys
import time
import traceback
from subprocess import (
    CalledProcessError,
    check_call as check_call_orig,
    PIPE,
    Popen,
    )


class IntegrationTestBase:

    banner_width = 70
    test_type = None
    juju_target = None
    default_juju_env = None

    def __init__(self):
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument(
            '--no-tear-down', action='store_true', default=False,
            help='Do not run the tear down method for debugging.')

        self.parser.add_argument(
            '-e', '--environment', default=self.default_juju_env,
            help='Juju environment to use as defined in '
                 '~/.juju/environment.yaml')
        self.args = self.parser.parse_args()
        self.juju_env = self.args.environment

    def banner(self, msg, width=banner_width):
        print '*' * width
        print '* ' + msg.ljust(width - 4) + ' *'
        print '*' * width

    def check_call(self, *args):
        print '-' * 70
        print "check_call:"
        print ' '.join(*args)
        print '-' * 70
        try:
            check_call_orig(*args)
        except CalledProcessError as e:
            print '*' * 70
            print e.output
            print '*' * 70
            raise

    def on_remote(self, args):
        if isinstance(args, basestring):
            args = args.split()

        self.check_call(
            'juju ssh {} -o StrictHostKeyChecking=no'
            ' -o UserKnownHostsFile=/dev/null'
            ' -e {}'.format(self.juju_target, self.juju_env).split()
            + list(args))

    def bzr_whoami(self):
        addr = formataddr(['Not A Real Person', 'no@example.com'])
        self.on_remote('bzr whoami "{}"'.format(addr))

    def set_up(self):
        """Set up a juju-managed instance to run the tests on."""
        self.banner(
            'Setting up the test environment for {}.'.format(self.test_type))
        self.check_call('juju bootstrap -e {}'.format(self.juju_env).split())
        # XXX The "ubuntu" charm is broken, so it has to be loaded from the
        # local repository.  Get it from
        # lp:~charmers/charms/precise/ubuntu/trunk.
        self.check_call(
            'juju deploy local:ubuntu --repository ~/juju-charms'
            ' -e {}'.format(self.juju_env).split())

        start_time = time.time()
        while time.time() - start_time < 600:
            process = Popen(
                'juju status -e {}'.format(self.juju_env).split(), stdout=PIPE)
            stdout, stderr = process.communicate()
            if 'instance-state: running' in stdout:
                break
        else:
            raise RuntimeError('starting the instance took too long')

        # Even though the instance-state is "running", it's still not ready, so
        # wait a little while before continuing.
        time.sleep(30)

        self.on_remote('sudo apt-add-repository --yes ppa:yellow/ppa')
        self.on_remote('sudo apt-get update')
        self.on_remote('sudo apt-get install python-shelltoolbox')
        self.check_call(
            'juju scp -o StrictHostKeyChecking=no'
            ' -o UserKnownHostsFile=/dev/null -e {}'
            ' -r . {}:lpsetup'.format(self.juju_env, self.juju_target).split())

    def tear_down(self):
        self.banner('Cleaning up.')
        code = os.system(
            'echo y | juju destroy-environment -e {}'.format(self.juju_env))
        # Provide a line-feed to follow the previous command.
        print
        if code != 0:
            raise RuntimeError('Destroying the test juju environment failed.')

    def check_environment(self):
        self.banner('Checking test environment.')
        code = os.system('juju status -e {}'.format(self.juju_env))
        if code == 0:
            # The "juju status" should have failed.
            raise RuntimeError('A juju environment unexpectedly exists.')

    def do_test(self):
        self.banner('Running integration tests for {}.'.format(self.test_type))
        self.bzr_whoami()

    def run(self):
        start_time = datetime.now()
        self.check_environment()
        try:
            self.set_up()
            try:
                self.do_test()
            except Exception as err:
                _, _, tb = sys.exc_info()
                self.banner('Test failed.  Sorry.')
                print err
                traceback.print_tb(tb)
                return 1
            else:
                self.banner('Test succeeded.')
                return 0
        finally:
            if not self.args.no_tear_down:
                self.tear_down()
            self.banner("Run time: {}".format(datetime.now() - start_time))
