#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the init-target subcommand."""

import os
import shutil
import stat
import tempfile
import unittest

from lpsetup.subcommands import init_target
from lpsetup.tests.utils import get_random_string


class WriteFileContentsTestCase(unittest.TestCase):
    """Tests for init_target.write_file_contents()."""

    def setUp(self):
        temp_file = tempfile.NamedTemporaryFile()
        # Calling close() on the temp_file will delete it, but its name
        # will live on...
        temp_file.close()
        self.temp_filename = temp_file.name

    def test_write_file_contents_writes_file_contents(self):
        # init_target.write_file_contents writes the supplied file contents
        # to a given file using the mode provided and chmods the file to
        # 0644 for the current user.
        self.addCleanup(os.remove, self.temp_filename)
        init_target.write_file_contents(
            self.temp_filename, 'Hello, world!', 'w')
        with open(self.temp_filename, 'r') as the_file:
            contents = the_file.read()
        self.assertEqual('Hello, world!', contents)
        file_mode = os.stat(self.temp_filename).st_mode
        # The file permissions should be 0644. So, we do a bitwise
        # dance...
        expected_mode = (
            stat.S_IWUSR | stat.S_IRUSR | stat.S_IRGRP | stat.S_IWGRP |
            stat.S_IROTH | ~ stat.S_IXUSR | ~ stat.S_IWGRP |
            ~ stat.S_IXGRP | ~ stat.S_IWOTH | ~ stat.S_IXOTH)
        self.assertTrue(file_mode ^ expected_mode)

    def test_write_file_contents_wont_overwrite_existing_files(self):
        # If the file already exists, `mode` is set to 'w', and the file's
        # contents differ from `contents`, write_file_contents() raises an
        # Exception.
        self.addCleanup(os.remove, self.temp_filename)
        init_target.write_file_contents(
            self.temp_filename, 'Hello, world!', 'w')
        self.assertRaises(
            Exception, init_target.write_file_contents, self.temp_filename,
            'Hello again!', 'w')

    def test_write_file_contents_returns_if_file_contents_dont_change(self):
        # If the existing file's contents don't differ from `contents`,
        # write_file_contents() will simply return.
        self.addCleanup(os.remove, self.temp_filename)
        init_target.write_file_contents(
            self.temp_filename, 'Hello, world!', 'w')
        # This second call shouldn't raise an exception.
        init_target.write_file_contents(
            self.temp_filename, 'Hello, world!', 'w')

    def test_write_file_contents_accepts_modes_a_or_w(self):
        # Passing a `mode` other than "a" or "w" to write_file_contents()
        # will cause an Exception.
        self.assertRaises(
            AssertionError, init_target.write_file_contents,
            '/tmp/foo', 'Hello, world!', 'a+')


class SetupSSHTest(unittest.TestCase):
    """Tests for init_target.setup_ssh()."""

    def setUp(self):
        temp_file = tempfile.NamedTemporaryFile()
        temp_file.close()
        self.temp_filename = temp_file.name
        self.ssh_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.ssh_dir)
        self.addCleanup(os.remove, self.temp_filename)
        self.addCleanup(os.remove, self.temp_filename + '.pub')

    def test_setup_ssh_generates_keys_if_not_present(self):
        # If SSH keys do not exist, the function will generate some.
        init_target.setup_ssh(self.ssh_dir, False, self.temp_filename)
        self.assertTrue(os.path.exists(self.temp_filename))
        self.assertTrue(os.path.exists(self.temp_filename + '.pub'))

    def test_setup_ssh_generates_authorized_keys(self):
        # setup_ssh() also generates an authorized_keys file containing
        # the public key.
        init_target.setup_ssh(self.ssh_dir, False, self.temp_filename)
        authorized_keys = os.path.join(self.ssh_dir, 'authorized_keys')
        self.assertTrue(os.path.exists(authorized_keys))
        public_key = open(self.temp_filename + '.pub').read()
        self.assertIn(public_key, open(authorized_keys).read())

    def test_setup_ssh_generates_known_hosts(self):
        # setup_ssh() also generates a `known_host` file containing
        # 'bazaar.launchpad.net'.
        init_target.setup_ssh(self.ssh_dir, False, self.temp_filename)
        known_hosts = os.path.join(self.ssh_dir, 'known_hosts')
        self.assertTrue(os.path.exists(known_hosts))
        self.assertIn('bazaar.launchpad.net', open(known_hosts).read())

    def test_setup_ssh_creates_ssh_dir(self):
        # If the ssh_dir passed to setup_ssh() doesn't exist, it
        # will be created.
        shutil.rmtree(self.ssh_dir)
        init_target.setup_ssh(self.ssh_dir, False, self.temp_filename)
        self.assertTrue(os.path.exists(self.ssh_dir))
        self.assertTrue(os.path.isdir(self.ssh_dir))


class VersionTests(unittest.TestCase):
    """Tests for version and upgrade data storage."""

    def setUp(self):
        self.directories_made = []

    def mkdirs(self, directory):
        self.directories_made.append(directory)

    def test_version_dir_creation(self):
        # If the version path does not exist, it is created.
        init_target.make_version_dir(mkdirs=self.mkdirs)
        self.assertIn(init_target.get_version_path(), self.directories_made)

    def test_originals_dir_creation(self):
        # If the version path contains an "originals" directory for backups of
        # files that we modify.
        init_target.make_version_dir(mkdirs=self.mkdirs)
        self.assertIn(init_target.get_version_path(), self.directories_made)


def create(filename, contents=None):
    """Create an empty file."""
    f = open(filename, 'a')
    if contents is not None:
        f.write(contents)
    f.close()


class BackupTests(unittest.TestCase):
    """Make sure the backup functions work correctly."""

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)

    def test_backups(self):
        # When asked to make a backup, make_backup does.
        source = os.path.join(self.tempdir, 'source')
        destination_dir = lambda: self.tempdir
        contents = get_random_string()
        create(source, contents)
        init_target.make_backup(source, destination_dir)
        get_backup_path = destination_dir() + os.path.sep + source
        self.assertTrue(os.path.exists(get_backup_path))
        self.assertEqual(file(get_backup_path).read(), contents)
