#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the argparser module."""

import argparse
from collections import OrderedDict
import itertools
import sys
import textwrap
import unittest

from lpsetup import argparser
from lpsetup.utils import get_terminal_width
from lpsetup.exceptions import ValidationError
from lpsetup.tests.utils import (
    call_replaced_by,
    ParserTestMixin,
    StepsFactoryTestMixin,
    SubCommandFactoryTestMixin,
    )


class AddCommonArgumentsTest(ParserTestMixin, unittest.TestCase):

    def test_interactive_supported(self):
        # Ensure the resulting *namespace.interactive* is True if the sub
        # command is interactive and `--yes` is not provided.
        argparser.add_common_arguments(self.parser, has_interactive_run=True)
        namespace = self.parser.parse_args([])
        self.assertTrue(namespace.interactive)

    def test_interactive_not_supported(self):
        # Ensure the resulting *namespace.interactive* is False if the sub
        # command does not support interactive execution.
        argparser.add_common_arguments(self.parser, has_interactive_run=False)
        namespace = self.parser.parse_args([])
        self.assertFalse(namespace.interactive)

    def test_assume_yes_provided(self):
        # Ensure the resulting *namespace.interactive* is False if the sub
        # command is interactive but `--yes` is provided.
        argparser.add_common_arguments(self.parser, has_interactive_run=True)
        namespace = self.parser.parse_args(['--yes'])
        self.assertFalse(namespace.interactive)

    def test_dry_run_supported(self):
        # Ensure the resulting *namespace.dry_run* is False if the sub
        # command supports dry run execution but `--dry-run` is not provided.
        argparser.add_common_arguments(self.parser, has_dry_run=True)
        namespace = self.parser.parse_args([])
        self.assertFalse(namespace.dry_run)

    def test_dry_run_not_supported(self):
        # Ensure the resulting *namespace.dry_run* is False if the sub
        # command does not support dry run execution.
        argparser.add_common_arguments(self.parser, has_dry_run=False)
        namespace = self.parser.parse_args([])
        self.assertFalse(namespace.dry_run)

    def test_dry_run_provided(self):
        # Ensure the resulting *namespace.dry_run* is True if the sub
        # command supports dry run execution and `--dry-run` is provided.
        argparser.add_common_arguments(self.parser, has_dry_run=True)
        namespace = self.parser.parse_args(['--dry-run'])
        self.assertTrue(namespace.dry_run)


class ArgumentParserTest(ParserTestMixin, unittest.TestCase):

    def test_add_argument(self):
        # Ensure actions are stored in a "public" instance attribute.
        self.parser.add_argument('arg1')
        self.parser.add_argument('arg2')
        dests = [action.dest for action in self.parser.actions]
        self.assertListEqual(['help', 'arg1', 'arg2'], dests)


class BaseSubCommandTest(SubCommandFactoryTestMixin, unittest.TestCase):

    def setUp(self):
        self.namespace = argparse.Namespace()

    def _successful_handler(self, namespace):
        namespace.bar = True

    def _failing_handler(self, namespace):
        raise ValidationError('nothing is going on')

    def test_get_description(self):
        # The default subcommand description is an empty string.
        subcommand = self.make_subcommand()
        self.assertEqual('', subcommand.get_description(self.namespace))

    def test_prepare_namespace(self):
        # The handlers are executed when *subcommand.prepare_namespace*
        # is invoked.
        subcommand = self.make_subcommand(handlers=[self._successful_handler])
        subcommand.prepare_namespace(self.namespace)
        self.assertTrue(self.namespace.bar)

    def test_failing_handler(self):
        # Ensure a validation error is raised by *subcommand.prepare_namespace*
        # if the namespace is not valid.
        subcommand = self.make_subcommand(handlers=[self._failing_handler])
        with self.assertRaises(ValidationError) as cm:
            subcommand.prepare_namespace(self.namespace)
        self.assertIn('nothing is going on', str(cm.exception))


class GetArgsFromNamespaceTest(ParserTestMixin, unittest.TestCase):

    def test_different_namespace(self):
        # It is possible to recreate the argument list taking values from
        # a different namespace.
        self.parser.add_argument('--foo')
        self.parser.add_argument('bar')
        namespace = self.parser.parse_args('--foo eggs spam'.split())
        namespace.foo = 'changed'
        args = argparser.get_args_from_namespace(self.parser, namespace)
        self.assertSequenceEqual(['--foo', 'changed', 'spam'], args)

    def test_with_multiple_values(self):
        # Ensure *get_args_from_namespace* correctly handles options
        # accepting multiple values.
        self.parser.add_argument('foo')
        self.parser.add_argument('--bar', nargs='+')
        namespace = self.parser.parse_args('foo --bar eggs spam'.split())
        namespace.bar.append('another argument')
        args = argparser.get_args_from_namespace(self.parser, namespace)
        expected = ['foo', '--bar', 'eggs', 'spam', 'another argument']
        self.assertSequenceEqual(expected, args)

    def test_boolean_values(self):
        # Ensure *get_args_from_namespace* correctly handles options
        # accepting boolean values.
        self.parser.add_argument('--foo', action='store_true')
        self.parser.add_argument('--bar', action='store_false')
        expected = ['--foo', '--bar']
        namespace = self.parser.parse_args(expected)
        args = argparser.get_args_from_namespace(self.parser, namespace)
        self.assertSequenceEqual(expected, args)


class GetStepNameTest(StepsFactoryTestMixin, unittest.TestCase):

    def test_named_step(self):
        # Ensure the *step_name* attribute of a step, if present, is returned.
        step = self.make_named_step('mystep')
        self.assertEqual('mystep', argparser.get_step_name(step))

    def test_function(self):
        # If the *step_name* attribute is not present, *__name__* is used.
        step = self.make_step()
        self.assertEqual(step.__name__, argparser.get_step_name(step))


class GetStepDescriptionTest(StepsFactoryTestMixin, unittest.TestCase):

    def assertStartsWith(self, prefix, content):
        """Assert that content starts with prefix."""
        self.assertTrue(
            content.startswith(prefix),
            '%r does not start with %r' % (content, prefix))

    def test_with_context(self):
        # Ensure the description is correctly retrieved and formatted.
        step = self.make_described_step('This step will do $stuff.')
        description = argparser.get_step_description(step, stuff='nothing')
        self.assertEqual('* This step will do nothing.', description)

    def test_without_context(self):
        # The description can be still retrieved if no context is provided.
        original = 'This step will do $stuff.'
        step = self.make_described_step(original)
        description = argparser.get_step_description(step)
        self.assertEqual('* ' + original, description)

    def test_without_placeholder(self):
        # Ensure an error is raised if a placeholder is missing.
        step = self.make_described_step('This step will do $stuff.')
        with self.assertRaises(KeyError):
            argparser.get_step_description(step, foo='bar')

    def test_missing_description(self):
        # Ensure an empty string is returned if the description is None.
        step = self.make_described_step(None)
        description = argparser.get_step_description(step)
        self.assertEqual('', description)

    def test_no_description(self):
        # Ensure an AttrubuteError is raised if the description is not found.
        with self.assertRaises(AttributeError):
            argparser.get_step_description(self.make_step())

    def test_dedent(self):
        # Ensure the description is correctly dedented.
        original = """
            Hi there!
        """
        step = self.make_described_step(original)
        description = argparser.get_step_description(step)
        self.assertEqual('* Hi there!', description)

    def test_empty_lines(self):
        # Ensure empty lines in description are removed.
        step = self.make_described_step('Hello.\n  \nGoodbye.')
        description = argparser.get_step_description(step)
        self.assertEqual('* Hello.\n* Goodbye.', description)

    def test_wrapping(self):
        # Ensure the description is correctly wrapped.
        width = get_terminal_width()
        elements = itertools.cycle('Lorem ipsum dolor sit amet.')
        original = ''.join(itertools.islice(elements, width + 1))
        step = self.make_described_step(original)
        description = argparser.get_step_description(step)
        lines = description.splitlines()
        self.assertEqual(2, len(lines))
        first_line, second_line = lines
        self.assertStartsWith('* ', first_line)
        self.assertStartsWith('  ', second_line)


class HelpFormatterTest(unittest.TestCase):

    def test_newlines(self):
        # Ensure the customized help formatter wraps text preserving new lines.
        formatter = argparser.HelpFormatter('prog', width=25)
        wrapped = 'This text will be wrapped.'
        preserved = '\nNew lines\nare preserved.'
        formatter.add_text(wrapped + preserved)
        expected = textwrap.fill(wrapped, 25) + preserved + '\n'
        self.assertEqual(expected, formatter.format_help())

    def test_tabs(self):
        # Ensure the customized help formatter wraps text preserving tabs.
        formatter = argparser.HelpFormatter('prog', width=80)
        text = '\tTabs are\tpreserved.'
        formatter.add_text(text)
        self.assertEqual(text.expandtabs() + '\n', formatter.format_help())

    def test_whitespaces(self):
        # Ensure the customized help formatter replaces multiple whitespaces
        # with a single whitespace, and drop leading and trailing whitespaces.
        formatter = argparser.HelpFormatter('prog', width=80)
        text = ' Text with leading, trailing and  multiple   whitespaces.  '
        formatter.add_text(text)
        expected = ' '.join(text.split()) + '\n'
        self.assertEqual(expected, formatter.format_help())


class InitNamespaceTest(unittest.TestCase):

    def test_user_info_in_namespace(self):
        # The namespace is initialized with current user info.
        namespace = argparse.Namespace()
        argparser.init_namespace(namespace)
        self.assertIsInstance(namespace.euid, int)
        self.assertIsInstance(namespace.run_as_root, bool)


class ResolveStepsTest(StepsFactoryTestMixin, unittest.TestCase):

    def setUp(self):
        self.step1 = self.make_step('step1')
        self.step2 = self.make_step('step2')
        self.step3 = self.make_named_step('step3')
        self.steps = [(self.step1, []), (self.step2, []), (self.step3, [])]

    def resolve(self, steps, namespace):
        """Resolve the steps and return them as an ordered dict."""
        return OrderedDict(argparser.resolve_steps(steps, namespace))

    def test_arguments_resolving(self):
        # Ensure the steps arguments are correctly resolved using
        # the current namespace.
        steps = [(self.step1, ['arg1', 'arg2']), (self.step2, ['arg3'])]
        namespace = argparse.Namespace(arg1='foo', arg2='bar', arg3='spam')
        steps_dict = self.resolve(steps, namespace)
        self.assertListEqual(['foo', 'bar'], steps_dict[self.step1])
        self.assertListEqual(['spam'], steps_dict[self.step2])

    def test_skip_steps(self):
        # Ensure steps are correctly skipped if requested.
        namespace = argparse.Namespace(skip_steps=['step1', 'step3'])
        steps = self.resolve(self.steps, namespace).keys()
        self.assertEqual(1, len(steps))
        self.assertIs(self.step2, steps[0])

    def test_include_steps(self):
        # Ensure not explicitly included steps are excluded.
        namespace = argparse.Namespace(steps=['step1', 'step3'])
        steps = self.resolve(self.steps, namespace).keys()
        self.assertListEqual([self.step1, self.step3], steps)

    def test_include_skipped_steps(self):
        # Ensure it's not possible to include a step already skipped.
        namespace = argparse.Namespace(
            steps=['step1'], skip_steps=['step1', 'step2', 'step3'])
        self.assertEqual({}, self.resolve(self.steps, namespace))

    def test_filter_function(self):
        # The filter function can grant or deny the step execution.
        steps = [
            (self.step1, [], lambda namespace: namespace.run_step1),
            (self.step2, [], lambda namespace: True)
            ]
        namespace = argparse.Namespace(run_step1=False)
        steps = self.resolve(steps, namespace).keys()
        self.assertEqual(1, len(steps))
        self.assertIs(self.step2, steps[0])


class RestartAsRootTest(ParserTestMixin, unittest.TestCase):

    def test_restart_normal(self):
        # Ensure an arbitrary command can be correctly restarted using `sudo`.
        args = ['--foo', 'bar']
        self.parser.add_argument(args[0])
        namespace = self.parser.parse_args(args)
        with call_replaced_by(lambda cmd: cmd):
            cmd = argparser.restart_as_root(self.parser, namespace)
        expected = ['sudo', sys.argv[0]] + args
        self.assertListEqual(expected, cmd)

    def test_restart_subcommand(self):
        # Ensure a subcommand is correctly restarted using `sudo`.
        name = 'subcommand-name'
        subparsers = self.parser.add_subparsers()
        subparser = subparsers.add_parser(name)
        namespace = self.parser.parse_args([name])
        with call_replaced_by(lambda cmd: cmd):
            cmd = argparser.restart_as_root(subparser, namespace)
        expected = ['sudo', sys.argv[0], name]
        self.assertListEqual(expected, cmd)


class StepsBasedSubCommandTest(
    SubCommandFactoryTestMixin, StepsFactoryTestMixin, unittest.TestCase):

    def make_subcommand_with_steps(self, *args):
        """Create a steps based subcommand containing given steps.

        Each step takes no arguments.
        """
        steps = [(step, []) for step in args]
        return self.make_steps_based_subcommand(steps=steps)

    def test_add_arguments(self):
        # The special arguments `--steps` and `--skip-steps` are
        # automatically added to the parser.
        step1 = self.make_named_step('step1')
        step2 = self.make_named_step('step2')
        step3 = self.make_named_step('step3')
        subcommand = self.make_subcommand_with_steps(step1, step2, step3)
        parser = argparse.ArgumentParser()
        subcommand.add_arguments(parser)
        steps_to_skip = ['step1', 'step2']
        steps_to_run = ['step3']
        namespace = parser.parse_args(
            ['--skip-steps'] + steps_to_skip + ['--steps'] + steps_to_run)
        self.assertListEqual(steps_to_skip, namespace.skip_steps)
        self.assertListEqual(steps_to_run, namespace.steps)

    def test_get_description(self):
        # Ensure the subcommand's description includes the description
        # of each step.
        step1 = self.make_described_step('step1 description')
        step2 = self.make_described_step(None)
        step3 = self.make_described_step('step3 description')
        subcommand = self.make_subcommand_with_steps(step1, step2, step3)
        description = subcommand.get_description(argparse.Namespace())
        # A step with no description is ignored.
        step1_description, step3_description = description.split('\n')
        self.assertIn(argparser.get_step_description(step1), step1_description)
        self.assertIn(argparser.get_step_description(step3), step3_description)

    def test_run(self):
        # Ensure the subcommand *run* method executes the steps
        # in the order they are provided.
        executed = []
        # Each step adds its name to *executed* when is run.
        step1 = lambda: executed.append('step1')
        step2 = lambda: executed.append('step2')
        step3 = lambda: executed.append('step3')
        subcommand = self.make_subcommand_with_steps(step1, step2, step3)
        subcommand.run(argparse.Namespace())
        self.assertListEqual(['step1', 'step2', 'step3'], executed)
