#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the handlers module."""

import argparse
from contextlib import contextmanager
import getpass
import os
import pwd
import shutil
import tempfile
import unittest

from shelltoolbox import (
    cd,
    environ,
    )

from lpsetup.exceptions import ValidationError
from lpsetup.handlers import (
    handle_branch_and_checkout,
    handle_directories,
    handle_lpuser_as_username,
    handle_lpuser_from_lplogin,
    handle_ssh_keys,
    handle_target_dir,
    handle_target_from_repository,
    handle_testing,
    handle_user,
    handle_userdata,
    )
from lpsetup.tests.utils import (
    lpuser,
    skip_if_no_lpuser,
    test_template_dir,
    )


class HandlersTestMixin(object):

    @contextmanager
    def assertNotValid(self, argument):
        try:
            yield
        except ValidationError as err:
            self.assertIn(argument, str(err))
        else:
            self.fail('ValidationError not raised')


class AssertNotValidTest(HandlersTestMixin, unittest.TestCase):

    def test_assertion(self):
        # Ensure the assertion does not fail if ValidationError is raised.
        with self.assertNotValid('error'):
            raise ValidationError('example error text')

    def test_assertion_fails_different_message(self):
        # Ensure the assertion fails if ValidationError is raised with
        # a different message.
        with self.assertRaises(AssertionError):
            with self.assertNotValid('failure'):
                raise ValidationError('example error text')

    def test_assertion_fails_no_exception(self):
        # Ensure the assertion fails if ValidationError is not raised.
        with self.assertRaises(AssertionError) as cm:
            with self.assertNotValid('error'):
                pass
        self.assertEqual('ValidationError not raised', str(cm.exception))

    def test_assertion_fails_different_exception(self):
        # Ensure the assertion function does not swallow other exceptions.
        with self.assertRaises(TypeError):
            with self.assertNotValid('error'):
                raise TypeError


class HandleBranchAndCheckoutTest(HandlersTestMixin, unittest.TestCase):

    def test_valid_arguments(self):
        # The validation succeed if there are no conflicts in arguments.
        namespace = argparse.Namespace(
            branch_name='branch',
            checkout_name='checkout')
        handle_branch_and_checkout(namespace)

    def test_checkout_equals_branch(self):
        # The validation fails if checkout and branch have the same name.
        namespace = argparse.Namespace(
            branch_name='branch',
            checkout_name='branch')
        with self.assertNotValid('checkout'):
            handle_branch_and_checkout(namespace)

    def test_checkout_equals_branch_without_trees(self):
        # The validation succeed if checkout and branch have the same name
        # but `--no-checkout` is specified.
        namespace = argparse.Namespace(
            branch_name='branch',
            checkout_name='branch',
            no_checkout=True)
        handle_branch_and_checkout(namespace)

    def test_invalid_names(self):
        # The validation fails if the branch/checkout name is '.' or '..',
        # or if it contains '/'.
        for kwargs in (
            {'branch_name': '.', 'checkout_name': 'checkout'},
            {'branch_name': 'branch', 'checkout_name': '..'},
            {'branch_name': '../', 'checkout_name': 'checkout'},
            {'branch_name': 'branch', 'checkout_name': 'check/out'},
            ):
            with self.assertNotValid('invalid name'):
                handle_branch_and_checkout(argparse.Namespace(**kwargs))


class HandleDirectoriesTest(HandlersTestMixin, unittest.TestCase):

    home_dir = os.path.expanduser('~')
    dependencies_dir = '~/launchpad/deps'

    def test_home_is_expanded(self):
        # Ensure the ~ construction is automatically expanded.
        namespace = argparse.Namespace(
            repository='~/launchpad', home_dir=self.home_dir,
            dependencies_dir=self.dependencies_dir)
        handle_directories(namespace)
        self.assertEqual(
            os.path.join(self.home_dir, 'launchpad'),
            namespace.repository)
        self.assertEqual(
            os.path.join(self.home_dir, 'launchpad', 'deps'),
            namespace.dependencies_dir)

    def test_tilde_expanded_only_when_needed(self):
        # Ensure the "~" construction is expanded only when it represents
        # the home directory of the user.
        repository = '~repository/~'
        namespace = argparse.Namespace(
            repository=repository, home_dir=self.home_dir)
        self.assertEqual(repository, namespace.repository)

    def test_directory_not_in_home(self):
        # The validation fails for directories not residing inside the home.
        namespace = argparse.Namespace(
            repository='/tmp/launchpad', home_dir=self.home_dir,
            dependencies_dir=self.dependencies_dir)
        with self.assertNotValid('repository'):
            handle_directories(namespace)

    def test_directory_containing_spaces(self):
        # The validation fails if the directory name contains spaces.
        namespace = argparse.Namespace(repository='directory with spaces')
        with self.assertNotValid('directory'):
            handle_directories(namespace)

    def test_relative_path(self):
        # Ensure relative paths are correctly expanded.
        namespace = argparse.Namespace(
            repository='launchpad', home_dir=self.home_dir,
            dependencies_dir='launchpad/../deps')
        with cd(self.home_dir):
            handle_directories(namespace)
            cwd = os.getcwd()
            self.assertEqual(
                os.path.join(cwd, 'launchpad'),
                namespace.repository)
            self.assertEqual(
                os.path.join(cwd, 'deps'),
                namespace.dependencies_dir)

    def test_invalid_relative_path(self):
        # The validation fails if a relative path points outside the home.
        namespace = argparse.Namespace(
            repository='~/../', home_dir=self.home_dir)
        with self.assertNotValid('repository'):
            handle_directories(namespace)


class HandleLPUserAsUsernameTest(unittest.TestCase):

    def test_lpuser(self):
        # If lpuser is not provided by namespace, the user name is used.
        username = getpass.getuser()
        namespace = argparse.Namespace(user=username, lpuser=None)
        handle_lpuser_as_username(namespace)
        self.assertEqual(username, namespace.lpuser)


class HandleLPUserFromLPLoginTest(HandlersTestMixin, unittest.TestCase):

    def temp_home(self):
        home = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, home)
        return environ(HOME=home)

    def test_lpuser_in_namespace(self):
        # Ensure nothing happens if the lpuser exists in the namespace.
        value = 'myuser'
        namespace = argparse.Namespace(lpuser=value)
        handle_lpuser_from_lplogin(namespace)
        self.assertEqual(value, namespace.lpuser)

    @skip_if_no_lpuser
    def test_lpuser_retrieval(self):
        # Ensure lpuser is correctly retrieved using bzr if it does not
        # exist in the namespace.
        namespace = argparse.Namespace()
        handle_lpuser_from_lplogin(namespace)
        self.assertEqual(lpuser, namespace.lpuser)

    def test_no_lpuser(self):
        # The validation fails if lpuser is not retrievable.
        with self.temp_home():
            namespace = argparse.Namespace()
            with self.assertNotValid('launchpad-login'):
                handle_lpuser_from_lplogin(namespace)

    def test_use_http(self):
        # Ensure lpuser is set to None if not retrievable and
        # *use_http* is True.
        with self.temp_home():
            namespace = argparse.Namespace(use_http=True)
            handle_lpuser_from_lplogin(namespace)
        self.assertIsNone(namespace.lpuser)


class HandleSSHKeysTest(HandlersTestMixin, unittest.TestCase):

    ssh_key_name = 'my_id_rsa'

    def setUp(self):
        self.home_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.home_dir)
        self.namespace = argparse.Namespace(
            ssh_key_name=self.ssh_key_name, home_dir=self.home_dir)

    def get_ssh_key_path(self):
        return os.path.join(self.home_dir, '.ssh', self.ssh_key_name)

    def write_key(self, path):
        directory = os.path.dirname(path)
        if not os.path.exists(directory):
            os.mkdir(os.path.dirname(path))
        with open(path, 'w') as f:
            f.write('KEY CONTENTS')

    def test_ssh_key_path_in_namespace(self):
        # After the handler is called, the ssh key path is present
        # as an attribute of the namespace.
        handle_ssh_keys(self.namespace)
        self.assertEqual(self.get_ssh_key_path(), self.namespace.ssh_key_path)

    def test_invalid_keys(self):
        # Keys are marked as invalid if they can not be found in the
        # current home directory.
        handle_ssh_keys(self.namespace)
        self.assertFalse(self.namespace.valid_ssh_keys)

    def test_valid_keys(self):
        # Keys are marked as valid if they are correctly found in the
        # current home directory.
        ssh_key_path = self.get_ssh_key_path()
        self.write_key(ssh_key_path)
        self.write_key(ssh_key_path + '.pub')
        handle_ssh_keys(self.namespace)
        self.assertTrue(self.namespace.valid_ssh_keys)

    def test_only_one_key(self):
        # Ensure the handler raises a ValidationError if only one key
        # of the private/public pair is found.
        ssh_key_path = self.get_ssh_key_path()
        self.write_key(ssh_key_path)
        with self.assertNotValid(ssh_key_path):
            handle_ssh_keys(self.namespace)


class HandleTargetDirTest(HandlersTestMixin, unittest.TestCase):

    def setUp(self):
        # Create a base target dir structure.
        self.target_dir = tempfile.mktemp()
        self.addCleanup(shutil.rmtree, self.target_dir)
        shutil.copytree(test_template_dir, self.target_dir)
        self.dirname = os.path.split(self.target_dir)[1]

    def test_absolute_path(self):
        # Ensure the resulting target dir is an absolute path.
        namespace = argparse.Namespace(target_dir=self.dirname)
        with cd(tempfile.tempdir):
            handle_target_dir(namespace)
        self.assertEqual(self.target_dir, namespace.target_dir)

    def test_expanduser(self):
        # Ensure the user is correctly expanded in the resulting target dir.
        namespace = argparse.Namespace(target_dir='~/' + self.dirname)
        with environ(HOME=tempfile.tempdir):
            handle_target_dir(namespace)
        self.assertEqual(self.target_dir, namespace.target_dir)

    def test_isdir(self):
        # An error is raised if the provided target dir is not a directory.
        with tempfile.NamedTemporaryFile() as f:
            namespace = argparse.Namespace(target_dir=f.name)
            with self.assertNotValid('is not a directory'):
                handle_target_dir(namespace)

    def test_exists(self):
        # An error is raised if the provided target dir does not exist.
        target_dir = os.path.join(self.target_dir, '__does_not_exist__')
        namespace = argparse.Namespace(target_dir=target_dir)
        with self.assertNotValid('does not exist'):
            handle_target_dir(namespace)

    def test_is_writable(self):
        # An error is raised if the target dir can not be written by
        # current user.
        namespace = argparse.Namespace(target_dir=self.target_dir)
        os.chmod(self.target_dir, 0500)
        self.addCleanup(os.chmod, self.target_dir, 0700)
        with self.assertNotValid('is not writable'):
            handle_target_dir(namespace)

    def test_is_launchpad_tree(self):
        # An error is raised if the target dir is not a valid Launchpad tree.
        os.remove(
            os.path.join(self.target_dir, 'utilities', 'update-sourcecode'))
        namespace = argparse.Namespace(target_dir=self.target_dir)
        with self.assertNotValid('is not a valid Launchpad tree'):
            handle_target_dir(namespace)


class HandleTargetFromRepositoryTest(unittest.TestCase):

    branch_name = 'branch'
    checkout_name = 'checkout'
    repository = '/repository/'

    def get_target_dir(self, no_checkout):
        namespace = argparse.Namespace(
            branch_name=self.branch_name,
            checkout_name=self.checkout_name,
            repository=self.repository,
            no_checkout=no_checkout)
        handle_target_from_repository(namespace)
        return namespace.target_dir

    def test_checkout(self):
        # Ensure the target dir points to branch_name if
        # lightweight checkout is not used.
        expected = os.path.join(self.repository, self.checkout_name)
        target_dir = self.get_target_dir(False)
        self.assertEqual(expected, target_dir)

    def test_no_checkout(self):
        # Ensure the target dir points to checkout_name if
        # lightweight checkout is used.
        expected = os.path.join(self.repository, self.branch_name)
        target_dir = self.get_target_dir(True)
        self.assertEqual(expected, target_dir)


class HandleTestingTest(unittest.TestCase):

    ctx = {
        'create_scripts': True,
        'install_haveged': False,
        'install_subunit': False,
        'no_checkout': True,
        'stop_lxc': False,
        'use_http': False,
        }

    def test_true(self):
        # Ensure aliased options are set to True if testing is True.
        namespace = argparse.Namespace(
            testing=True, interactive=True, **self.ctx)
        handle_testing(namespace)
        for key in self.ctx:
            self.assertTrue(getattr(namespace, key))
        self.assertFalse(namespace.interactive)

    def test_false(self):
        # Ensure no changes are made to aliased options if testing is False.
        namespace = argparse.Namespace(
            testing=False, interactive=True, **self.ctx)
        handle_testing(namespace)
        for key, value in self.ctx.items():
            self.assertEqual(value, getattr(namespace, key))
        self.assertTrue(namespace.interactive)


class HandleUserTest(HandlersTestMixin, unittest.TestCase):

    def test_home_dir(self):
        # Ensure validator populates namespace with `home_dir` name.
        username = getpass.getuser()
        namespace = argparse.Namespace(user=username)
        handle_user(namespace)
        self.assertEqual(pwd.getpwnam(username).pw_dir, namespace.home_dir)

    def test_failure(self):
        # Ensure the validation fails if the current user is root and
        # no user is provided.
        namespace = argparse.Namespace(user=None, euid=0)
        with self.assertNotValid('user'):
            handle_user(namespace)


class HandleUserDataTest(HandlersTestMixin, unittest.TestCase):

    def test_no_data(self):
        # If full name and email are not provided, the handler tries to
        # obtain them using the given `whois` callable.
        namespace = argparse.Namespace(
            full_name=None, email=None, user='root')
        email = 'email@example.com'
        handle_userdata(namespace, lambda user: (user, email))
        self.assertEqual(namespace.user, namespace.full_name)
        self.assertEqual(email, namespace.email)

    def test_no_whois(self):
        # Ensure the validation fails if full name or email are not provided
        # and they can not be obtained using the `whois` callable.
        namespace = argparse.Namespace(
            full_name=None, email=None, user='root')
        with self.assertNotValid('full-name'):
            handle_userdata(namespace, lambda user: None)

    def test_non_existent_user(self):
        # Ensure the validation fails if full name or email are not provided
        # and the system user does not exist.
        namespace = argparse.Namespace(
            full_name=None, email=None, user='this_user_does_not_exist')
        with self.assertNotValid('full-name'):
            handle_userdata(namespace, lambda user: None)

    def test_only_one_argument(self):
        # It does not make sense to provide only one argument.
        namespace = argparse.Namespace(full_name='Foo Bar', email=None)
        with self.assertNotValid('full-name'):
            handle_userdata(namespace, lambda user: None)
