#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""A collection of helper functions."""

__metaclass__ = type
__all__ = [
    'call',
    'confirm',
    'get_container_path',
    'get_file_header',
    'get_lxc_gateway',
    'get_network_interfaces',
    'get_running_containers',
    'get_terminal_width',
    'lxc_in_state',
    'lxc_ip',
    'lxc_stopped',
    'render_to_file',
    'retry',
    'running_in_container',
    'Scrubber',
    'sshlxc',
    'this_command',
    'wrap_contents',
    ]

from ConfigParser import RawConfigParser
from datetime import datetime
from functools import (
    partial,
    wraps,
    )
import errno
import glob
import os
import re
import subprocess
import shutil
import sys
import textwrap
import time
from uuid import uuid4

from shelltoolbox import (
    command,
    join_command,
    run,
    ssh,
    )

from lpsetup import (
    exceptions,
    get_version,
    )
from lpsetup.settings import (
    LXC_LP_DIR_PATTERN,
    LXC_LP_TEST_DIR_PATTERN,
    LXC_IP_ARGS,
    LXC_IP_COMMAND,
    LXC_PATH,
    )


PID_RE = re.compile("pid:\s+(\d+)")


call = partial(run, stdout=None)


class ConfigParser(RawConfigParser):
    """A customized configuration parser.

    The base `RawConfigParser` separates options from values using a colon
    or a equal sign. This parser forces the separator to be an equal sign,
    allowing the colon to be part of an option name.
    """
    OPTCRE = re.compile(
        r'(?P<option>[^=\s][^=]*)'  # very permissive!
        r'\s*(?P<vi>=)\s*'          # space/tab, separator, space/tab
        r'(?P<value>.*)$'           # everything up to eol
        )


def confirm():
    """Ask a yes/no confirmation to proceed.

    Return True if the answer is 'yes', False otherwise.
    When the user presses Enter without typing anything, False is assumed.
    """
    while True:
        response = raw_input('Do you want to proceed? [y/N] ').lower()
        if not response:
            return False
        if response in ('y', 'yes'):
            return True
        if response in ('n', 'no'):
            return False
        print("I didn't understand you. Please specify '(y)es' or '(n)o'.")


def get_container_path(lxc_name, path='', base_path=LXC_PATH):
    """Return the path of LXC container called `lxc_name`.

    If a `path` is given, return that path inside the container, e.g.::

        >>> get_container_path('mycontainer')
        '/var/lib/lxc/mycontainer/rootfs/'
        >>> get_container_path('mycontainer', '/etc/apt/')
        '/var/lib/lxc/mycontainer/rootfs/etc/apt/'
        >>> get_container_path('mycontainer', 'home')
        '/var/lib/lxc/mycontainer/rootfs/home'
    """
    return os.path.join(base_path, lxc_name, 'rootfs', path.lstrip('/'))


def get_file_header(program=sys.argv[0], now=None,
                    initial_indent='# ', subsequent_indent='# ',
                    version=None):
    if now is None:
        now = datetime.utcnow()
    if version is None:
        version = get_version()
    header = (
        'This file modified at {date}UTC by {program}, '
        'version {version}.'.format(
            date=datetime.isoformat(now),
            program=program,
            version=version,
            )
        )
    return textwrap.fill(
        header, initial_indent=initial_indent,
        subsequent_indent=subsequent_indent)


def wrap_contents(content, program=sys.argv[0], version=None, now=None,
                  uuid=None, initial_indent='# ', subsequent_indent='# '):
    """Given some content, wrap the content with a fence.

    The fence will consist of opening and closing markers and a header
    specifying the program name and time the content was made.

        >>> now = datetime(1969, 7, 20, 20, 18)
        >>> content = 'how now\\nbrown cow'
        >>> print wrap_contents(content, '/usr/bin/tester', '1.2',
        ...                     now=now, uuid='DEAD-BEEF')
        # --------------------/usr/bin/tester v1.2--------------------DEAD-BEEF
        # This file modified at 1969-07-20T20:18:00UTC by /usr/bin/tester,
        # version 1.2.
        how now
        brown cow
        # --------------------/usr/bin/tester v1.2--------------------DEAD-BEEF
        <BLANKLINE>
    """
    if now is None:
        now = datetime.utcnow()
    if uuid is None:
        uuid = str(uuid4())
    if version is None:
        version = get_version()
    prog_version = "{0} v{1}".format(program, version)
    open_marker = initial_indent + '-' * 20 + prog_version + '-' * 20 + uuid
    close_marker = initial_indent + '-' * 20 + prog_version + '-' * 20 + uuid

    header = get_file_header(
        program, now, subsequent_indent, subsequent_indent, version)
    template = textwrap.dedent("""\
    {open_marker}
    {header}
    {content}
    {close_marker}
    """)
    return template.format(open_marker=open_marker,
                           header=header,
                           content=content,
                           close_marker=close_marker)


def get_lxc_gateway(candidates=('lxcbr0', 'virbr0'), interface_lookup=None):
    """Return the gateway bridge name to be used by LXC containers.

    The gateway name will change depending on which version of Ubuntu the
    script is running on: if *lxcbr0* does not exist, *virbr0* will be used.
    If both interfaces don't exist, None is returned.
    """
    if interface_lookup is None:
        interfaces = get_network_interfaces()
    else:
        interfaces = interface_lookup()
    for interface in candidates:
        if interface in interfaces:
            return interface


def get_network_interfaces(path='/sys/class/net/', exclude=()):
    """Return a list of the network interfaces up in this system."""
    return [
        i for i in os.listdir(path)
        if i not in exclude and
        os.path.isdir(os.path.join(path, i))]


def get_running_containers(containers=None):
    """Return an iterable of currently running LXC container's names."""
    if containers is None:
        containers = run('lxc-ls').split()
    visited = {}
    for container in containers:
        if container in visited:
            yield container
        else:
            visited[container] = 1


def get_terminal_width():
    """Return the terminal width (number of columns).

    By default, `stty` is used to retrieve the terminal width.
    However, users can override the returned value exporting $COLUMNS
    (usually not exported).
    """
    default = 79
    # Try using the environment variable $COLUMNS.
    try:
        width = int(os.getenv('COLUMNS'))
    except (TypeError, ValueError):
        # $COLUMNS not exported or not a number: try using `stty`.
        try:
            width = int(run('stty', 'size').split()[1])
        except subprocess.CalledProcessError:
            # No input available for stty, return a default value.
            return default
    return width if width > 0 else default


def lxc_in_state(state, lxc_name, timeout=30):
    """Return True if the LXC named `lxc_name` is in state `state`.

    Return False otherwise.
    """
    while timeout:
        try:
            output = run('lxc-info', '-n', lxc_name, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            pass
        else:
            if state in output:
                return True
        timeout -= 1
        time.sleep(1)
    return False


lxc_stopped = partial(lxc_in_state, 'STOPPED')


def render_to_file(template_name, context, dest, templates_dir=None):
    """Render `template_name` using `context`. Write the result inside `dest`.

    The argument `template_dir` is the directory containing the template:
    if None, the `templates` directory inside the lpsetup package is used.

    Return the rendered contents.
    """
    if templates_dir is None:
        templates_dir = os.path.join(os.path.dirname(__file__), 'templates')
    with open(os.path.join(templates_dir, template_name)) as source:
        template = source.read()
    contents = template.format(**context)
    with open(dest, 'w') as destination:
        destination.write(contents)
    return contents


def retry(exception, tries=60, delay=0.5):
    """If the decorated function raises `exception`, wait and try it again.

    Raise the exception raised by the last call if the function does not
    exit normally after 100 tries.

    Original from http://wiki.python.org/moin/PythonDecoratorLibrary#Retry.
    """
    def decorator(func):
        @wraps(func)
        def decorated(*args, **kwargs):
            mtries = tries
            while mtries:
                try:
                    return func(*args, **kwargs)
                except exception as err:
                    time.sleep(delay)
                    mtries -= 1
            raise err
        return decorated
    return decorator


def running_in_container():
    """Are we running in a container?"""
    # 'running-in-container' is Precise and greater; 'lxc-is-container' is
    # Lucid.  These are provided by the lxcguest package.
    for cmd in ('running-in-container', 'lxc-is-container'):
        try:
            return not subprocess.call([cmd])
        except OSError, err:
            # ENOENT means "No such file or
            # directory."
            if err.errno != errno.ENOENT:
                raise
    # If neither of the commands exist we are not running in a container.
    # See http://tinyurl.com/running-in-container for details.
    return False


class Scrubber(object):
    """Scrubber will cleanup after lxc ephemeral uncleanliness.

    All running containers will be killed, except for the ones listed
    in the `ignored_containers` iterable.

    Those directories corresponding the lp_test_dir_pattern will
    be unmounted and removed.  The 'ephemeralbind' subdirectories
    will be unmounted.

    The directories corresponding to the lxc_lp_dir_pattern will
    be unmounted and removed.  No subdirectories will need
    unmounting.
    """
    def __init__(self, user, lxc_name,
                 lp_test_dir_pattern=LXC_LP_TEST_DIR_PATTERN,
                 lxc_lp_dir_pattern=LXC_LP_DIR_PATTERN,
                 ignored_containers=None):
        if ignored_containers is None:
            self.ignored_containers = []
        else:
            self.ignored_containers = list(ignored_containers)
        self.lp_test_dir_pattern = lp_test_dir_pattern.format(
            lxc_name=lxc_name)
        self.lxc_lp_dir_pattern = lxc_lp_dir_pattern
        self.user = user

    def umount(self, dir_):
        if os.path.ismount(dir_):
            run("umount", dir_)

    def scrubdir(self, dir_, extra):
        dirs = [dir_]
        if extra is not None:
            dirs.insert(0, os.path.join(dir_, extra))
        for d in dirs:
            self.umount(d)
        shutil.rmtree(dir_)

    def scrub(self, pattern, extra=None):
        for dir_ in glob.glob(pattern):
            if os.path.isdir(dir_):
                self.scrubdir(dir_, extra)

    def get_pid(self, container):
        info = run("lxc-info", "-n", container)
        # lxc-info returns a string containing 'RUNNING' for those
        # containers that are running followed by 'pid: <pid>', so
        # that must be parsed.
        if 'RUNNING' in info:
            match = PID_RE.search(info)
            if match:
                return match.group(1)
        return None

    def get_running_containers(self):
        """Get the running containers.

        Returns a list of (name, pid) tuples.
        """
        output = run("lxc-ls")
        containers = set(output.split()).difference(self.ignored_containers)
        pidlist = [(c, self.get_pid(c)) for c in containers]
        return [(c, p) for c, p in pidlist if p is not None]

    def killer(self):
        """Kill all running ephemeral containers."""
        pids = self.get_running_containers()
        if len(pids) > 0:
            # We can do this the easy way...
            for name, pid in pids:
                run("/usr/bin/lxc-stop", "-n", name)
            time.sleep(2)
            pids = self.get_running_containers()
            # ...or, the hard way.
            for name, pid in pids:
                run("kill", "-9", pid)

    def run(self):
        self.killer()
        self.scrub(self.lp_test_dir_pattern, "ephemeralbind")
        self.scrub(self.lxc_lp_dir_pattern)


@retry(subprocess.CalledProcessError)
def lxc_ip(name):
    """Return the ip address of the LXC container called `name`.

    This function will try to obtain the ip address of the container for
    ~30 seconds before raising a `subprocess.CalledProcessError`.
    """
    args = LXC_IP_COMMAND + LXC_IP_ARGS
    cmd = command(*args)
    return cmd(name).strip()


def sshlxc(name, cmd, **kwargs):
    """Run ssh shell commands.

    The ssh connection is made resolving the ip address of the LXC
    container called `name`.

    Raise `lpsetup.exceptions.ExecutionError` if the ip address is not found.
    Raise `subprocess.CalledProcessError` if the ssh command returns an error.
    """
    try:
        ip = lxc_ip(name)
    except subprocess.CalledProcessError as err:
        msg = 'Error: unable to find the ip address of the LXC container: '
        raise exceptions.ExecutionError(msg + str(err))
    sshcall = ssh(ip, **kwargs)
    return sshcall(cmd)


def this_command(directory, args, cmd=None):
    """Return a command line to run a command using given `args`.

    The returned command will execute the command from `directory`.
    If cmd is not specified, the invoking script is used::

        >>> script_name = sys.argv[0]
        >>> cmd = this_command('/home/user/lp/branches', ['install'])
        >>> cmd == 'cd /home/user/lp/branches && ' + script_name + ' install'
        True

        >>> script_name = 'ls'
        >>> cmd = this_command('/home/user/lp/branches',
        ...                    ['install'], script_name)
        >>> cmd == 'cd /home/user/lp/branches && ' + script_name + ' install'
        True

    Arguments are correctly quoted::

        >>> script_name = sys.argv[0]
        >>> cmd = this_command('/path', ['-arg', 'quote me'])
        >>> cmd == ('cd /path && ' + script_name + " -arg 'quote me'")
        True
    """
    if cmd is None:
        cmd = sys.argv[:1]
    if isinstance(cmd, basestring):
        cmd = [cmd]
    script = join_command(cmd + list(args))
    return join_command(['cd', directory]) + ' && ' + script
